/* PNAT.  Simple Port Trigger */
#include <linux/types.h>
#include <linux/ip.h>
#include <linux/timer.h>
#include <linux/module.h>
#include <linux/netfilter.h>
#include <linux/netdevice.h>
#include <linux/if.h>
#include <linux/skbuff.h>
#include <linux/udp.h>
#include <linux/inetdevice.h>
#include <net/protocol.h>
#include <net/checksum.h>
#include <linux/proc_fs.h>
#include <linux/netfilter_ipv4.h>
#include <linux/netfilter_ipv4/ip_nat_rule.h>
#include <linux/netfilter_ipv4/ipt_MARK.h>
#include "ipt_PNAT.h"

#if 0
#define DEBUGP printk
#else
#define DEBUGP(format, args...)
#endif

#define TIMEOUT	300 //Second

static DEFINE_RWLOCK(pnat_lock);
struct ipt_pnat_info pnat_info[MAX_PORT_TRIGGER];

static unsigned long get_expire(const struct ip_conntrack *conntrack)
{
	if(!conntrack)
		return 0;
	DEBUGP("get_expire\n");
	return (timer_pending(&conntrack->timeout) ? (conntrack->timeout.expires - jiffies)/HZ : 0);
}

void add_out_ct_list(struct ipt_pnat_info *info,struct ip_conntrack *ct)
{
	struct ip_conntrack_list *list=NULL;
	struct ip_conntrack_list *last=NULL;
	
	list=info->out_ct_list;
	while(list){
		last=list;
		list=list->next;
	}
	if(last==NULL){
		info->out_ct_list=kmalloc(sizeof(struct ip_conntrack_list), GFP_ATOMIC);
		info->out_ct_list->ct=ct;
		info->out_ct_list->next=NULL;
	}else{
		last->next=kmalloc(sizeof(struct ip_conntrack_list), GFP_ATOMIC);
		last->next->ct=ct;
		last->next->next=NULL;
	}
}

void add_in_ct_list(struct ipt_pnat_info *info,struct ip_conntrack *ct)
{
	struct ip_conntrack_list *list=NULL;
	struct ip_conntrack_list *last=NULL;
	
	list=info->in_ct_list;
	while(list){
		last=list;
		list=list->next;
	}
	if(last==NULL){
		info->in_ct_list=kmalloc(sizeof(struct ip_conntrack_list), GFP_ATOMIC);
		info->in_ct_list->ct=ct;
		info->in_ct_list->next=NULL;
	}else{
		last->next=kmalloc(sizeof(struct ip_conntrack_list), GFP_ATOMIC);
		last->next->ct=ct;
		last->next->next=NULL;
	}
}

/* check is timeout or not */
static unsigned long check_expire(struct ipt_pnat_info *info)
{
	struct ip_conntrack_list *prev=NULL;
	struct ip_conntrack_list *list=NULL;
	unsigned long expire_time=0;
	unsigned long max_expire_time=0;
	
	write_lock_bh(&pnat_lock);
	list=info->out_ct_list;
	while(list){
		expire_time=get_expire(list->ct);
		DEBUGP(">>>OUT: ID=%d , expire_time=%ld\n",list->ct->id,expire_time);
		if(max_expire_time < expire_time)
			max_expire_time = expire_time;
		
		/* remove unnecessary entry */	
		if(expire_time <= 0 || list->ct->id <= 0){
			//first
			if(list==info->out_ct_list)
			{
				info->out_ct_list=list->next;
				kfree(list);
			}
			else
			{
				prev->next=list->next;
				kfree(list);
				list=prev;
			}
		}
		prev=list;
		list=list->next;
	}
	DEBUGP(">>>OUT: max_expire_time=%ld\n",max_expire_time);
	prev=NULL;
	list=info->in_ct_list;
	while(list){
		expire_time=get_expire(list->ct);
		DEBUGP(">>>IN: ID=%d , expire_time=%ld\n",list->ct->id, expire_time);
		if(max_expire_time < expire_time)
			max_expire_time = expire_time;
		
		/* remove unnecessary entry */	
		if(expire_time <= 0 || list->ct->id <= 0){
			//first
			if(list==info->in_ct_list)
			{
				info->in_ct_list=list->next;
				kfree(list);
			}
			else
			{
				prev->next=list->next;
				kfree(list);
				list=prev;
			}
		}
		prev=list;
		list=list->next;
	}
	DEBUGP(">>>FINAL: max_expire_time=%ld\n\n",max_expire_time);
	write_unlock_bh(&pnat_lock);
	
	if(info->timeout == 0)
	return max_expire_time;
	else
		return max_expire_time > info->timeout ? info->timeout : max_expire_time;
	
}

/* free ct_list */
void free_out_ct_list(struct ipt_pnat_info *info)
{
	struct ip_conntrack_list *list=NULL;
	struct ip_conntrack_list *next=NULL;
	
	list=info->out_ct_list;
	while(list){
		next=list->next;
		kfree(list);
		list=next;
	}
}

void free_in_ct_list(struct ipt_pnat_info *info)
{
	struct ip_conntrack_list *list=NULL;
	struct ip_conntrack_list *next=NULL;
	
	list=info->in_ct_list;
	while(list){
		next=list->next;
		kfree(list);
		list=next;
	}
}

/* find outbound ifname proto and port range */
int find_outbound(struct ipt_pnat_info **info_p,char *ifname,u_int16_t protocol,u_int16_t port,u_int32_t pkt_ip)
{
	int i=0, ret=0;
	for(i=0;i<MAX_PORT_TRIGGER;i++){
		if(pnat_info[i].outbound.ifname[0]==0)	
			continue;
		/* not use */
		if(pnat_info[i].outbound.min==0)
			continue;

		if(strcmp(pnat_info[i].outbound.ifname,ifname)!=0)
			continue;

		//any or match
		if(pkt_ip == pnat_info[i].config_ip || pnat_info[i].config_ip == 0)
		{
		/* tcp or both*/
		if((protocol==IPPROTO_TCP && strcmp(pnat_info[i].outbound.protocol,"udp")!=0)
			/* udp or both */
			||( protocol==IPPROTO_UDP && strcmp(pnat_info[i].outbound.protocol,"tcp")!=0)){
			
				if(port >= pnat_info[i].outbound.min 
							&& port <= pnat_info[i].outbound.max){
						info_p[i] = &pnat_info[i];
						ret = 1;
		}
	}
	return NULL;
}
	}
	return ret;
}

/* find inbound ifname proto and port range */
struct ipt_pnat_info* find_inbound(char *ifname,u_int16_t protocol,u_int16_t port)
{
	int i=0;
	
	for(i=0;i<MAX_PORT_TRIGGER;i++){
		if(pnat_info[i].inbound.ifname[0]==0)	
			continue;
		/* not use */
		if(pnat_info[i].inbound.min==0)
			continue;
		/* compare name */
		if(strcmp(pnat_info[i].inbound.ifname,ifname)!=0)
			continue;

		DEBUGP("in port=%u %u %u\n",port,pnat_info[i].inbound.min,pnat_info[i].inbound.max);			
		/* tcp or both*/
		if((protocol==IPPROTO_TCP && strcmp(pnat_info[i].inbound.protocol,"udp")!=0)
			/* udp or both */
			||( protocol==IPPROTO_UDP && strcmp(pnat_info[i].inbound.protocol,"tcp")!=0)){
			
				if(port >= pnat_info[i].inbound.min 
						&& port <= pnat_info[i].inbound.max)
					return &pnat_info[i];
		}
	}
	return NULL;
}

/* FIXME: Take multiple ranges --RR */
static int pnat_check(const char *tablename,
		const void *entry,
		const struct xt_target *target,
		void *targinfo,
		unsigned int targinfosize,
		unsigned int hook_mask)
{

	if (strcmp(tablename, "nat") != 0) {
		DEBUGP("pnat_check: bad table `%s'.\n", tablename);
		return 0;
	}
	if (targinfosize != IPT_ALIGN(sizeof(struct ipt_mark_target_info))) {
		DEBUGP("pnat_check: size %u.\n", targinfosize);
		return 0;
	}
	if (hook_mask & ~((1 << NF_IP_PRE_ROUTING))) {
		DEBUGP("pnat_check: bad hooks %x.\n", hook_mask);
		return 0;
	}
	return 1;
}

static unsigned int pnat_target(struct sk_buff **pskb,
		const struct net_device *in,
		const struct net_device *out,
		unsigned int hooknum,
		const struct xt_target *target,
		const void *targinfo,
		void *userdata)
{
	enum ip_conntrack_info ctinfo;
	const struct ipt_mark_target_info *markinfo = targinfo;
//	struct ip_nat_multi_range newrange;
	struct ip_nat_range newrange;
	struct iphdr *iph = (*pskb)->nh.iph;
	void *protoh = (u_int32_t *)iph + iph->ihl;
	u_int16_t port;	
	struct ip_conntrack *ct;
	struct ipt_pnat_info *pnat;
	char name[16];
	struct ipt_pnat_info *pnat_p[MAX_PORT_TRIGGER];
	int i, ret;
		
	memset(pnat_p, 0, sizeof(pnat_p));
	ct = ip_conntrack_get(*pskb, &ctinfo);
	
	IP_NF_ASSERT(ct && (ctinfo == IP_CT_NEW || ctinfo == IP_CT_RELATED));
	
	
	IP_NF_ASSERT(hooknum == NF_IP_PRE_ROUTING || hooknum == NF_IP_LOCAL_OUT);
	
	if(hooknum==NF_IP_LOCAL_OUT)
		return IPT_CONTINUE;	

	if(!in->name)
		return IPT_CONTINUE;
	
	strcpy(name,in->name);
	
	switch (iph->protocol){
		case IPPROTO_TCP:
		{
			struct tcphdr *tcph = protoh;
			port=ntohs(tcph->dest);
			break;
		}
			
		case IPPROTO_UDP:
		{
			struct udphdr *udph = protoh;
			port=ntohs(udph->dest);
			
			break;
		 }
			
		default:
			/* only support udp and tcp */
			return IPT_CONTINUE;
	}
	
	/* LAN to WAN is outbound */
	ret = find_outbound(pnat_p,name,iph->protocol,port,iph->saddr);
	
	for(i=0; i<MAX_PORT_TRIGGER; i++)
	{
		if(pnat_p[i])
		{	
			DEBUGP("OUT : iph->saddr=%u.%u.%u.%u port=%d\n",NIPQUAD(iph->saddr), port);
			// timeout ,init again , config timeout==0 ,init always
			if(check_expire(pnat_p[i])==0 || pnat_p[i]->timeout == 0)
			{
				pnat_p[i]->trigger_ip = pnat_p[i]->config_ip;
				free_out_ct_list(pnat_p[i]);
				//free_in_ct_list(pnat_p[i]);
			}
			/* server is anyone and never trigger */
			if(pnat_p[i]->trigger_ip == 0){
				pnat_p[i]->trigger_ip=iph->saddr;
			}
			if(iph->saddr == pnat_p[i]->trigger_ip)
			{
				ct->sc_timeout = pnat_p[i]->timeout;
				add_out_ct_list(pnat_p[i],ct);
			}
		}
	}
	if(ret)
		return IPT_CONTINUE;
	
	/* WAN to LAN is inbound */
	pnat=find_inbound(name,iph->protocol,port);
	
	if(pnat){
		DEBUGP("IN : iph->saddr=%u.%u.%u.%u port=%d\n",NIPQUAD(iph->saddr),port);
		/* expired stop DNAT */	
		if(check_expire(pnat)==0){
			//pnat->trigger_ip=0;
			free_in_ct_list(pnat);
			//free_out_ct_list(pnat);
			return IPT_CONTINUE;
		}
		ct->sc_timeout = pnat->timeout;
		add_in_ct_list(pnat,ct);
		/* Transfer from original range. */
		newrange = ((struct ip_nat_range)
		 	{ IP_NAT_RANGE_MAP_IPS,
				pnat->trigger_ip, 
				pnat->trigger_ip,
				{0}, 
				{0} });
		/* do MARK for FORWARD Policy is DROP */
	        if((*pskb)->nfmark != markinfo->mark) {
			(*pskb)->nfmark = markinfo->mark;
	//		(*pskb)->nfcache |= NFC_ALTERED;
	        }
		/* Hand modified range to generic setup. */
		return ip_nat_setup_info(ct, &newrange, hooknum);
	}
	return IPT_CONTINUE;
}

static struct ipt_target pnat_reg = {
		.name           = "PNAT",
		.family         = AF_INET,
		.target         = pnat_target,
		.targetsize     = sizeof(struct ipt_mark_target_info),
		.table          = "nat",
		.hooks          = (1<< NF_IP_PRE_ROUTING),
		.checkentry     = pnat_check,
		.me             = THIS_MODULE };

static ssize_t proc_read_pnat_fops(struct file *filp,
                                 char *buf,size_t count , loff_t *offp)
{
	char tmp[1024];
	int len=0;
	int i=0;
	//struct ipt_pnat_info *pnat;
	if(*offp!=0)
		return 0;
	
	for(i=0;i < MAX_PORT_TRIGGER;i++){
		if(pnat_info[i].outbound.min!=0 
				&& pnat_info[i].outbound.max!=0
				&& pnat_info[i].inbound.min!=0
				&& pnat_info[i].inbound.max!=0){
			/*
			pnat=find_outbound(pnat_info[i].outbound.ifname
					,(*pnat_info[i].outbound.protocol=='t')?IPPROTO_TCP:IPPROTO_UDP
					,pnat_info[i].outbound.min);*/
			len+=sprintf(tmp+len,"%s %s [%d-%d] %s %s [%d-%d] trigger ip %u.%u.%u.%u expire time %ld\n"
				,pnat_info[i].outbound.ifname
				,pnat_info[i].outbound.protocol
				,pnat_info[i].outbound.min
				,pnat_info[i].outbound.max
				,pnat_info[i].inbound.ifname
				,pnat_info[i].inbound.protocol
				,pnat_info[i].inbound.min
				,pnat_info[i].inbound.max
				,NIPQUAD(pnat_info[i].trigger_ip)
				,check_expire(&pnat_info[i]));
		}
	}
	
	copy_to_user(buf,tmp,len);
	*offp = len;
	return len;
}

static ssize_t proc_write_pnat_fops(struct file *filp,const char *buffer,
                                     size_t count , loff_t *offp)
{
	char line[128];
	int i=0;
	
	
	copy_from_user(line,buffer,count);
	line[count]=0x00;
	
	for(i=0;i < MAX_PORT_TRIGGER;i++){
		if(pnat_info[i].outbound.min==0 
				|| pnat_info[i].outbound.max==0
				|| pnat_info[i].inbound.min==0
				|| pnat_info[i].inbound.max==0)
			break;
		/* clean table */
		if(line[0]=='0')
		{
			free_out_ct_list(&pnat_info[i]);
			free_in_ct_list(&pnat_info[i]);
		}
	}
	
	/* clean table */
	if(line[0]=='0'){
		memset(&pnat_info , 0 , sizeof(struct ipt_pnat_info) * MAX_PORT_TRIGGER);
		printk("Clean\n");
		return count;
	}
		
	
		
	if(i< MAX_PORT_TRIGGER){
		sscanf(line,"%s %s %d-%d %s %s %d-%d %u %ld"
				,pnat_info[i].outbound.ifname
				,pnat_info[i].outbound.protocol
				,(int *)&pnat_info[i].outbound.min
				,(int *)&pnat_info[i].outbound.max
				,pnat_info[i].inbound.ifname
				,pnat_info[i].inbound.protocol
				,(int *)&pnat_info[i].inbound.min
				,(int *)&pnat_info[i].inbound.max
				,(u_int32_t *)&pnat_info[i].config_ip
				,(unsigned long *)&pnat_info[i].timeout);
		
		// Assign timeout
		pnat_info[i].timeout = TIMEOUT;
		// Accept Any
		pnat_info[i].config_ip = 0;		
		pnat_info[i].trigger_ip = pnat_info[i].config_ip;
		
		if(pnat_info[i].outbound.min==0 
				|| pnat_info[i].outbound.max==0
				|| pnat_info[i].inbound.min==0
				|| pnat_info[i].inbound.max==0){
			printk("Invaild format\n");
			printk("lan_if proto min port-max port  wan_if proto min port-max port trigger ip timeout\n");
		}
		
	}else
		printk("Only support %d entry\n",MAX_PORT_TRIGGER);

	return count;
};	

static struct file_operations pnat_fops = {
        read: proc_read_pnat_fops,
        write: proc_write_pnat_fops,
};

static struct proc_dir_entry *proc_special;
static struct proc_dir_entry *proc_pnat;

static int __init init(void)
{
	
	proc_special=create_proc_entry("special", S_IFDIR, &proc_root);
	
	if (proc_special == NULL) {
		printk("create_proc_entry() failed\n");
		return -1;
        }
	proc_pnat=create_proc_entry("pnat",0666,proc_special);
	proc_pnat->owner = THIS_MODULE;
	proc_pnat->proc_fops = &pnat_fops;
	
	/* init pnat_info */	
	memset(&pnat_info , 0 , sizeof(struct ipt_pnat_info) * MAX_PORT_TRIGGER);
	
	return ipt_register_target(&pnat_reg);
}

static void __exit fini(void)
{
	int i=0;
	
	remove_proc_entry("pnat",proc_special);
	remove_proc_entry("special",&proc_root);
	
	/* free table */
	for(i=0;i < MAX_PORT_TRIGGER;i++){
		free_out_ct_list(&pnat_info[i]);
		free_in_ct_list(&pnat_info[i]);
	}

	ipt_unregister_target(&pnat_reg);
}

module_init(init);
module_exit(fini);
MODULE_LICENSE("GPL");
